import malaya

def test_to_cardinal():
    assert malaya.to_cardinal(123456789) == 'seratus dua puluh tiga juta empat ratus lima puluh enam ribu tujuh ratus lapan puluh sembilan'

def test_to_cardinal_negative():
    assert 'negatif' in malaya.to_cardinal(-123456789)

def test_to_cardinal_negative_decimal():
    assert 'perpuluhan' in malaya.to_cardinal(-1234567.89)

def test_to_ordinal():
    assert malaya.to_ordinal(11) == 'kesebelas'

def test_to_ordinal_belas():
    assert malaya.to_ordinal(12) == 'kedua belas'

def test_to_ordinal_one():
    try:
        malaya.to_ordinal(1.1)
    except:
        pass
    assert malaya.to_ordinal(1) == 'pertama'

def test_to_cardinal_one():
    assert malaya.to_cardinal(1) == 'satu'

def test_to_cardinal_belas():
    assert malaya.to_cardinal(12) == 'dua belas'

def test_to_ordinal_num():
    assert malaya.to_ordinal_num(11) == 'ke-11'

def test_to_currency():
    assert malaya.to_currency(123456789) == 'seratus dua puluh tiga juta empat ratus lima puluh enam ribu tujuh ratus lapan puluh sembilan ringgit'

def test_to_year():
    assert malaya.to_year(123456789) == 'seratus dua puluh tiga juta empat ratus lima puluh enam ribu tujuh ratus lapan puluh sembilan'



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from itertools import combinations
from toolz import compose
import random
from sklearn.feature_extraction.text import CountVectorizer


class SkipGramVectorizer(CountVectorizer):
    def __init__(self, skip = 1, **kwds):
        super(SkipGramVectorizer, self).__init__(**kwds)
        self.skip = skip

    def build_sent_analyzer(self, preprocess, stop_words, tokenize):
        return lambda sent: self._word_skip_grams(
            compose(tokenize, preprocess, self.decode)(sent), stop_words
        )

    def build_analyzer(self):
        preprocess = self.build_preprocessor()
        stop_words = self.get_stop_words()
        tokenize = self.build_tokenizer()
        sent_analyze = self.build_sent_analyzer(
            preprocess, stop_words, tokenize
        )

        return lambda doc: self._sent_skip_grams(doc, sent_analyze)

    def _sent_skip_grams(self, doc, sent_analyze):
        skip_grams = []
        for sent in doc.split('\n'):
            skip_grams.extend(sent_analyze(sent))
        return skip_grams

    def _word_skip_grams(self, tokens, stop_words = None):
        if stop_words is not None:
            tokens = [w for w in tokens if w not in stop_words]
        min_n, max_n = self.ngram_range
        skip = self.skip
        if max_n != 1:
            original_tokens = tokens
            if min_n == 1:
                tokens = list(original_tokens)
                min_n += 1
            else:
                tokens = []

            n_original_tokens = len(original_tokens)
            tokens_append = tokens.append
            space_join = ' '.join

            for n in range(min_n, min(max_n + 1, n_original_tokens + 1)):
                for i in range(n_original_tokens - n + 1):
                    head = [original_tokens[i]]
                    for skip_tail in combinations(
                        original_tokens[i + 1 : i + n + skip], n - 1
                    ):
                        tokens_append(space_join(head + list(skip_tail)))
        return tokens


def features_crf(sentence, index):
    return {
        'word': sentence[index],
        'is_first': index == 0,
        'is_last': index == len(sentence) - 1,
        'prefix-1': sentence[index][0],
        'prefix-2': sentence[index][:2],
        'prefix-3': sentence[index][:3],
        'suffix-1': sentence[index][-1],
        'suffix-2': sentence[index][-2:],
        'suffix-3': sentence[index][-3:],
        'prev_word': '' if index == 0 else sentence[index - 1],
        'prev_word-prefix-1': '' if index == 0 else sentence[index - 1][0],
        'prev_word-prefix-2': '' if index == 0 else sentence[index - 1][:2],
        'prev_word-prefix-3': '' if index == 0 else sentence[index - 1][:3],
        'prev_word-suffix-1': '' if index == 0 else sentence[index - 1][-1],
        'prev_word-suffix-2': '' if index == 0 else sentence[index - 1][-2:],
        'prev_word-suffix-3': '' if index == 0 else sentence[index - 1][-3:],
        'next_word-prefix-1': ''
        if index == len(sentence) - 1
        else sentence[index + 1][0],
        'next_word-prefix-2': ''
        if index == len(sentence) - 1
        else sentence[index + 1][:2],
        'next_word-prefix-3': ''
        if index == len(sentence) - 1
        else sentence[index + 1][:3],
        'next_word-suffix-1': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-1],
        'next_word-suffix-2': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-2:],
        'next_word-suffix-3': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-3:],
        'next_word': '' if index == len(sentence) - 1 else sentence[index + 1],
        'has_hyphen': '-' in sentence[index],
        'is_numeric': sentence[index].isdigit(),
    }


def features_crf_dependency(sentence, tag, index):
    return {
        'word': sentence[index],
        'tag': tag[index],
        'is_first': index == 0,
        'is_last': index == len(sentence) - 1,
        'prefix-1': sentence[index][0],
        'prefix-2': sentence[index][:2],
        'prefix-3': sentence[index][:3],
        'suffix-1': sentence[index][-1],
        'suffix-2': sentence[index][-2:],
        'suffix-3': sentence[index][-3:],
        'prev_word': '' if index == 0 else sentence[index - 1],
        'prev_word-prefix-1': '' if index == 0 else sentence[index - 1][0],
        'prev_word-prefix-2': '' if index == 0 else sentence[index - 1][:2],
        'prev_word-prefix-3': '' if index == 0 else sentence[index - 1][:3],
        'prev_word-suffix-1': '' if index == 0 else sentence[index - 1][-1],
        'prev_word-suffix-2': '' if index == 0 else sentence[index - 1][-2:],
        'prev_word-suffix-3': '' if index == 0 else sentence[index - 1][-3:],
        'next_word-prefix-1': ''
        if index == len(sentence) - 1
        else sentence[index + 1][0],
        'next_word-prefix-2': ''
        if index == len(sentence) - 1
        else sentence[index + 1][:2],
        'next_word-prefix-3': ''
        if index == len(sentence) - 1
        else sentence[index + 1][:3],
        'next_word-suffix-1': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-1],
        'next_word-suffix-2': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-2:],
        'next_word-suffix-3': ''
        if index == len(sentence) - 1
        else sentence[index + 1][-3:],
        'next_word': '' if index == len(sentence) - 1 else sentence[index + 1],
        'has_hyphen': '-' in sentence[index],
        'is_numeric': sentence[index].isdigit(),
    }


def skipgrams(
    sequence,
    vocabulary_size,
    window_size = 4,
    negative_samples = 1.0,
    shuffle = True,
    categorical = False,
    sampling_table = None,
    seed = None,
):
    couples = []
    labels = []
    for i, wi in enumerate(sequence):
        if not wi:
            continue
        if sampling_table is not None:
            if sampling_table[wi] < random.random():
                continue

        window_start = max(0, i - window_size)
        window_end = min(len(sequence), i + window_size + 1)
        for j in range(window_start, window_end):
            if j != i:
                wj = sequence[j]
                if not wj:
                    continue
                couples.append([wi, wj])
                if categorical:
                    labels.append([0, 1])
                else:
                    labels.append(1)

    if negative_samples > 0:
        num_negative_samples = int(len(labels) * negative_samples)
        words = [c[0] for c in couples]
        random.shuffle(words)

        couples += [
            [words[i % len(words)], random.randint(1, vocabulary_size - 1)]
            for i in range(num_negative_samples)
        ]
        if categorical:
            labels += [[1, 0]] * num_negative_samples
        else:
            labels += [0] * num_negative_samples

    if shuffle:
        if seed is None:
            seed = random.randint(0, 10e6)
        random.seed(seed)
        random.shuffle(couples)
        random.seed(seed)
        random.shuffle(labels)

    return couples, labels



import malaya

isu_kerajaan = ['Institusi raja khususnya Yang di-Pertuan Agong adalah kedaulatan negara dengan kedudukan dan peranannya termaktub dalam Perlembagaan Persekutuan yang perlu disokong dan didukung oleh kerajaan serta rakyat.',
               'Pensyarah Kulliyah Undang-Undang Ahmad Ibrahim, Universiti Islam Antarabangsa Malaysia (UIAM) Prof Madya Dr Shamrahayu Ab Aziz berkata perubahan kerajaan, susulan kemenangan Pakatan Harapan pada Pilihan Raya Umum Ke-14 pada Mei lepas, tidak memberi kesan dari segi peranan, fungsi dan kedudukan Yang di-Pertuan Agong.',
               'Peralihan kerajaan itu menyaksikan Sultan Muhammad V mencatat sejarah tersendiri dengan menjadi Yang di-Pertuan Agong Malaysia yang pertama memerintah dalam era dua kerajaan berbeza.',
               'Semasa dilantik sebagai Yang di-Pertuan Agong Ke-15 pada 13 Dis 2016, kerajaan ketika itu diterajui oleh Barisan Nasional dan pada 10 Mei lepas, kepimpinan negara diambil alih oleh Pakatan Harapan yang memenangi Pilihan Raya Umum Ke-14.',
               'Ketika merasmikan Istiadat Pembukaan Penggal Pertama, Parlimen ke-14 pada 17 Julai lepas, Seri Paduka bertitah mengalu-alukan pendekatan kerajaan Pakatan Harapan dalam menegakkan ketelusan terutamanya dengan mendedahkan kedudukan kewangan negara yang sebenar serta mengkaji semula perbelanjaan, kos projek dan mengurus kewangan secara berhemat bagi menangani kos sara hidup.',
               'Pada Jun lepas, Sultan Muhammad V memperkenankan supaya peruntukan gaji dan emolumen Yang di-Pertuan Agong dikurangkan sebanyak 10 peratus sepanjang pemerintahan sehingga 2021 berikutan keprihatinan Seri Paduka terhadap tahap hutang dan keadaan ekonomi negara.',
               'Seri Paduka turut menitahkan supaya Majlis Rumah Terbuka Aidilfitri tahun ini tidak diadakan di Istana Negara dengan peruntukan majlis itu digunakan bagi membantu golongan yang kurang bernasib baik.']

def test_lsa():
    assert len(malaya.summarize_lsa(isu_kerajaan)['top-words'])

def test_lsa_original():
    assert len(malaya.summarize_lsa(isu_kerajaan,maintain_original=True)['top-words'])

def test_lsa_not_cluster():
    assert len(malaya.summarize_lsa(isu_kerajaan,return_cluster=False)['top-words'])

def test_nmf_not_cluster():
    assert len(malaya.summarize_nmf(isu_kerajaan,return_cluster=False)['top-words'])

def test_lda_not_cluster():
    assert len(malaya.summarize_lda(isu_kerajaan,return_cluster=False)['top-words'])

def test_nmf():
    assert len(malaya.summarize_nmf(isu_kerajaan)['top-words'])

def test_lda():
    assert len(malaya.summarize_lda(isu_kerajaan)['top-words'])



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
from ._utils._utils import check_file, check_available
from ._models._sklearn_model import LANGUAGE_DETECTION
from ._models._tensorflow_model import DEEP_LANG
from ._utils._paths import PATH_LANG_DETECTION, S3_PATH_LANG_DETECTION

lang_labels = {0: 'OTHER', 1: 'ENGLISH', 2: 'INDONESIA', 3: 'MALAY'}


def label():
    """
    Return language labels dictionary.
    """
    return lang_labels


def multinomial(validate = True):
    """
    Load multinomial language detection model.
    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    LANGUAGE_DETECTION : malaya._models._sklearn_model.LANGUAGE_DETECTION class
    """
    if validate:
        check_file(
            PATH_LANG_DETECTION['multinomial'],
            S3_PATH_LANG_DETECTION['multinomial'],
        )
    else:
        if not check_available(PATH_LANG_DETECTION['multinomial']):
            raise Exception(
                'language-detection/multinomial is not available, please `validate = True`'
            )
    try:
        with open(PATH_LANG_DETECTION['multinomial']['vector'], 'rb') as fopen:
            vector = pickle.load(fopen)
        with open(PATH_LANG_DETECTION['multinomial']['model'], 'rb') as fopen:
            model = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('language-detection/multinomial') and try again"
        )
    return LANGUAGE_DETECTION(model, lang_labels, vector)


def sgd(validate = True):
    """
    Load SGD language detection model.
    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    LANGUAGE_DETECTION : malaya._models._sklearn_model.LANGUAGE_DETECTION class
    """
    if validate:
        check_file(PATH_LANG_DETECTION['sgd'], S3_PATH_LANG_DETECTION['sgd'])
    else:
        if not check_available(PATH_LANG_DETECTION['sgd']):
            raise Exception(
                'language-detection/sgd is not available, please `validate = True`'
            )
    try:
        with open(PATH_LANG_DETECTION['sgd']['vector'], 'rb') as fopen:
            vector = pickle.load(fopen)
        with open(PATH_LANG_DETECTION['sgd']['model'], 'rb') as fopen:
            model = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('language-detection/sgd') and try again"
        )
    return LANGUAGE_DETECTION(model, lang_labels, vector)


def xgb(validate = True):
    """
    Load XGB language detection model.
    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    LANGUAGE_DETECTION : malaya._models._sklearn_model.LANGUAGE_DETECTION class
    """
    if validate:
        check_file(PATH_LANG_DETECTION['xgb'], S3_PATH_LANG_DETECTION['xgb'])
    else:
        if not check_available(PATH_LANG_DETECTION['xgb']):
            raise Exception(
                'language-detection/xgb is not available, please `validate = True`'
            )
    try:
        with open(PATH_LANG_DETECTION['xgb']['vector'], 'rb') as fopen:
            vector = pickle.load(fopen)
        with open(PATH_LANG_DETECTION['xgb']['model'], 'rb') as fopen:
            model = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('language-detection/xgb') and try again"
        )
    return LANGUAGE_DETECTION(model, lang_labels, vector, mode = 'xgb')


def deep_model(validate = True):
    """
    Load deep learning language detection model.
    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    DEEP_LANG : malaya._models._tensorflow_model.DEEP_LANG class
    """
    import os

    if validate:
        check_file(PATH_LANG_DETECTION['deep'], S3_PATH_LANG_DETECTION['deep'])
    else:
        if not check_available(PATH_LANG_DETECTION['deep']):
            raise Exception(
                'language-detection/deep is not available, please `validate = True`'
            )
    try:
        with open(PATH_LANG_DETECTION['deep']['vector'], 'rb') as fopen:
            vector = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('language-detection/deep') and try again"
        )
    return DEEP_LANG(
        os.path.dirname(PATH_LANG_DETECTION['deep']['model']),
        vector,
        lang_labels,
    )



shortform = {
    'parti islam semalaysia': ['parti islam semalaysia', 'pas'],
    'parti keadilan rakyat': ['parti keadilan rakyat', 'pkr'],
    'parti amanah': ['parti amanah', 'pan'],
    'democratic action party': ['dap', 'democratic action party'],
    'Pilihan Raya Umum Ke-14': ['ge', 'pru', 'pilihan raya umum ke-14'],
    'malaysian chinese association': ['malaysian chinese association', 'mca'],
    'pakatan harapan': ['ph', 'pakatan harapan'],
    'barisan nasional': ['bn', 'barisan nasional'],
    'universiti malaya': ['universiti malaya', 'um'],
    'universiti islam antarabangsa': [
        'universiti islam antarabangsa',
        'uia',
        'iium',
    ],
    'maktab rendah sains mara': ['mjsc', 'mrsm', 'maktab rendah sains mara'],
}



from scipy.stats.mstats import gmean, hmean, hdmedian
import numpy as np


def _most_common(l):
    return max(set(l), key = l.count)


def voting_stack(models, text):
    """
    Stacking for POS, Entities and Dependency models.

    Parameters
    ----------
    models: list
        list of models
    text: str
        string to predict

    Returns
    -------
    result: list
    """
    if not isinstance(models, list):
        raise ValueError('models must be a list')
    if not isinstance(text, str):
        raise ValueError('text must be a string')
    results, texts, votes, votes_indices, indices = [], [], [], [], []
    is_dependency = False
    for i in range(len(models)):
        if not 'predict' in dir(models[i]):
            raise ValueError('all models must able to predict')
        predicted = models[i].predict(text)
        if isinstance(predicted, tuple):
            is_dependency = True
            predicted, indexing = predicted
            indexing = np.array(indexing)
            indices.append(indexing[:, 1:2])

        predicted = np.array(predicted)
        results.append(predicted[:, 1:2])
        texts.append(predicted[:, 0])
    concatenated = np.concatenate(results, axis = 1)
    for row in concatenated:
        votes.append(_most_common(row.tolist()))
    if is_dependency:
        concatenated = np.concatenate(indices, axis = 1)
        for row in concatenated:
            votes_indices.append(_most_common(row.tolist()))
    output = list(map(lambda X: (X[0], X[1]), list(zip(texts[-1], votes))))
    if is_dependency:
        return (
            output,
            list(
                map(lambda X: (X[0], X[1]), list(zip(texts[-1], votes_indices)))
            ),
        )
    else:
        return output


def predict_stack(models, text, mode = 'gmean'):
    """
    Stacking for predictive models.

    Parameters
    ----------
    models: list
        list of models
    text: str
        string to predict
    mode : str, optional (default='gmean')
        Model architecture supported. Allowed values:

        * ``'gmean'`` - geometrical mean
        * ``'hmean'`` - harmonic mean
        * ``'mean'`` - mean
        * ``'min'`` - min
        * ``'max'`` - max
        * ``'median'`` - Harrell-Davis median


    Returns
    -------
    result: dict
    """
    if not isinstance(models, list):
        raise ValueError('models must be a list')
    if not isinstance(text, str):
        raise ValueError('text must be a string')
    if not isinstance(mode, str):
        raise ValueError('mode must be a string')
    if mode.lower() == 'gmean':
        mode = gmean
    elif mode.lower() == 'hmean':
        mode = hmean
    elif mode.lower() == 'mean':
        mode = np.mean
    elif mode.lower() == 'min':
        mode = np.amin
    elif mode.lower() == 'max':
        mode = np.amax
    elif mode.lower() == 'median':
        mode = hdmedian
    else:
        raise Exception(
            "mode not supported, only support ['gmean','hmean','mean','min','max','median']"
        )
    labels, results = [], []
    for i in range(len(models)):
        nested_results = []
        if not 'predict' in dir(models[i]):
            raise ValueError('all models must able to predict')
        result = (
            models[i].predict(text)
            if models[i].predict.__defaults__ is None
            else models[i].predict(text, get_proba = True)
        )
        for key, item in result.items():
            if 'attention' in key:
                continue
            if key not in labels:
                labels.append(key)
            nested_results.append(item)
        results.append(nested_results)
    results = mode(np.array(results), axis = 0)
    return {label: results[no] for no, label in enumerate(labels)}



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from itertools import product
from collections import defaultdict, Counter
from fuzzywuzzy import fuzz
import numpy as np
import pulp
from scipy.spatial.distance import euclidean
from .texts._text_functions import normalizer_textcleaning
from .texts._tatabahasa import (
    rules_normalizer,
    consonants,
    vowels,
    sounds,
    GO,
    PAD,
    EOS,
    UNK,
)
from .spell import _return_possible, _edit_normalizer, _return_known
from .similarity import is_location


def _tokens_to_fracdict(tokens):
    cntdict = defaultdict(lambda: 0)
    for token in tokens:
        cntdict[token] += 1
    totalcnt = sum(cntdict.values())
    return {token: float(cnt) / totalcnt for token, cnt in cntdict.items()}


def _word_mover(left_token, right_token, vectorizer):
    all_tokens = list(set(left_token + right_token))
    wordvecs = {
        token: vectorizer.get_vector_by_name(token) for token in all_tokens
    }
    left_bucket = _tokens_to_fracdict(left_token)
    right_bucket = _tokens_to_fracdict(right_token)

    T = pulp.LpVariable.dicts(
        'T_matrix', list(product(all_tokens, all_tokens)), lowBound = 0
    )
    prob = pulp.LpProblem('WMD', sense = pulp.LpMinimize)
    prob += pulp.lpSum(
        [
            T[token1, token2] * euclidean(wordvecs[token1], wordvecs[token2])
            for token1, token2 in product(all_tokens, all_tokens)
        ]
    )
    for token2 in right_bucket:
        prob += (
            pulp.lpSum([T[token1, token2] for token1 in left_bucket])
            == right_bucket[token2]
        )
    for token1 in left_bucket:
        prob += (
            pulp.lpSum([T[token1, token2] for token2 in right_bucket])
            == left_bucket[token1]
        )
    prob.solve()
    return prob


def distance(left_token, right_token, vectorizer):
    """
    calculate word mover distance between left hand-side sentence and right hand-side sentence

    Parameters
    ----------
    left_token : list
        Eg, ['saya','suka','makan','ayam']
    right_token : list
        Eg, ['saya','suka','makan','ikan']
    vectorizer : object
        fast-text or word2vec interface object

    Returns
    -------
    distance: float
    """
    if not isinstance(left_token, list):
        raise ValueError('left_token must be a list')
    if not isinstance(right_token, list):
        raise ValueError('right_token must be a list')
    if not hasattr(vectorizer, 'get_vector_by_name'):
        raise ValueError('vectorizer must has `get_vector_by_name` method')
    prob = _word_mover(left_token, right_token, vectorizer)
    return pulp.value(prob.objective)


class _DEEP_CONTRACTION:
    def __init__(self, corpus, vectorizer):
        self.corpus = Counter(corpus)
        self.vectorizer = vectorizer

    def _suggest(self, string):
        """
        Normalize a string

        Parameters
        ----------
        string : str

        debug : bool, optional (default=True)
            If true, it will print character similarity distances.

        Returns
        -------
        string: normalized string
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        result, outer_candidates = [], []
        for word in normalizer_textcleaning(string).split():
            if word.istitle():
                result.append(word)
                continue
            if len(string) > 2:
                if string[-2] in consonants and string[-1] == 'e':
                    string[-1] = 'a'
            if word[0] == 'x' and len(word) > 1:
                result_string = 'tak '
                word = word[1:]
            else:
                result_string = ''
            if word[-2:] == 'la':
                end_result_string = ' lah'
                word = word[:-2]
            elif word[-3:] == 'lah':
                end_result_string = ' lah'
                word = word[:-3]
            else:
                end_result_string = ''
            if word in sounds:
                result.append(result_string + sounds[word] + end_result_string)
                continue
            if word in rules_normalizer:
                result.append(
                    result_string + rules_normalizer[word] + end_result_string
                )
                continue
            if word in self.corpus:
                result.append(result_string + word + end_result_string)
                continue
            candidates = (
                _return_known([word], self.corpus)
                or _return_known(_edit_normalizer(word), self.corpus)
                or _return_possible(word, self.corpus, _edit_normalizer)
                or [word]
            )
            candidates = list(candidates)
            candidates = [
                (candidate, is_location(candidate))
                for candidate in list(candidates)
            ]
            strings = [fuzz.ratio(string, k[0]) for k in candidates]
            descending_sort = np.argsort(strings)[::-1]
            selected = None
            for index in descending_sort:
                if not candidates[index][1]:
                    selected = candidates[index][0]
                    break
            selected = (
                candidates[descending_sort[0]][0] if not selected else selected
            )
            result.append(result_string + word + end_result_string)
            outer_candidates.append([(word, k[0]) for k in candidates])
        return ' '.join(result), outer_candidates

    def expand(self, string):
        result, candidates = self._suggest(string)
        intermediates = []
        for candidate in candidates:
            inner = []
            for c in candidate:
                text = [c[1] if w in c[0] else w for w in result.split()]
                inner.append(
                    (
                        c[0],
                        ' '.join(text),
                        distance(text, result.split(), self.vectorizer),
                    )
                )
            inner.sort(key = lambda x: x[2])
            intermediates.append(inner)
        return intermediates


def expander(corpus, vectorizer):
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not hasattr(vectorizer, 'get_vector_by_name'):
        raise ValueError('vectorizer must has `get_vector_by_name` method')
    return _DEEP_CONTRACTION(corpus, vectorizer)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
import os
import random
import numpy as np
from fuzzywuzzy import fuzz
import json
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from .texts.vectorizer import SkipGramVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.utils import shuffle
from .texts._text_functions import (
    STOPWORDS,
    STOPWORD_CALON,
    simple_textcleaning,
    str_idx,
)
from .generator import sentence_ngram
from . import home, _delete_macos
from ._utils._utils import download_file
from ._models._skip_thought import (
    train_model as skip_train,
    batch_sequence,
    load_skipthought as load_deep_skipthought,
)
from ._models._siamese_lstm import (
    train_model as siamese_train,
    load_siamese as load_deep_siamese,
)

from .topic import calon
from .topic import location


def _apply_stopwords_calon(string):
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string.lower()).strip()
    return ' '.join(
        [i for i in string.split() if i not in STOPWORD_CALON and len(i) > 1]
    )


_namacalon = [
    _apply_stopwords_calon(val['NamaCalon'].lower()) for _, val in calon.items()
]
_namacalon = list(set(_namacalon))
_negeri = list(set([val['negeri'].lower() for _, val in location.items()]))
_parlimen = list(set([val['parlimen'].lower() for _, val in location.items()]))
_dun = list(set([val['dun'].lower() for _, val in location.items()]))
_location = list(set(_negeri + _parlimen + _dun))


class _DEEP_SIAMESE_SIMILARITY:
    def __init__(
        self,
        sess,
        model,
        keys,
        dictionary,
        maxlen,
        saver,
        embedding_size,
        num_layers,
        output_size,
        dropout,
        is_influencers = False,
    ):
        self._sess = sess
        self._model = model
        self.keys = keys
        self.dictionary = dictionary
        self.maxlen = maxlen
        self._saver = saver
        self._embedding_size = embedding_size
        self._num_layers = num_layers
        self._output_size = output_size
        self._dropout = dropout

    def save_model(self, location):
        """
        save model to tensorflow checkpoint file.

        Parameters
        ----------
        string: str

        """
        if not isinstance(location, str):
            raise ValueError('location must be a string')
        self._saver.save(self._sess, '%s/model.ckpt' % (location))
        with open('%s/model.json' % (location), 'w') as fopen:
            json.dump(
                {
                    'dictionary': self.dictionary,
                    'maxlen': self.maxlen,
                    'keys': self.keys,
                    'embedding_size': self._embedding_size,
                    'num_layers': self._num_layers,
                    'output_size': self._output_size,
                    'dropout': self._dropout,
                },
                fopen,
            )

    def get_similarity(self, string, anchor = 0.5):
        """
        Return similar topics / influencers.

        Parameters
        ----------
        string: str
        anchor: float, (default=0.5)
            baseline similarity.

        Returns
        -------
        results: list of strings
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        if not isinstance(anchor, float):
            raise ValueError('anchor must be a float')
        if not (anchor > 0 and anchor < 1):
            raise ValueError('anchor must be bigger than 0, less than 1')
        original_string = simple_textcleaning(string)
        strings = [original_string] * len(self.keys)
        left = str_idx(strings, self.dictionary, self.maxlen, UNK = 3)
        right = str_idx(self.keys, self.dictionary, self.maxlen, UNK = 3)
        distances = self._sess.run(
            1 - self._model.distance,
            feed_dict = {self._model.X_left: left, self._model.X_right: right},
        )
        where = np.where(distances > anchor)[0]
        results = [self.keys[i].lower() for i in where]
        return list(set(results))


class _DEEP_SIMILARITY:
    def __init__(
        self,
        sess,
        model,
        vectorized,
        keys,
        dictionary,
        maxlen,
        saver,
        embedding_size,
        is_influencers = False,
    ):
        self._sess = sess
        self._model = model
        self.vectorized = vectorized
        self.keys = keys
        self.dictionary = dictionary
        self.maxlen = maxlen
        self._saver = saver
        self._embedding_size = embedding_size

    def save_model(self, location):
        """
        save model to tensorflow checkpoint file.

        Parameters
        ----------
        string: str

        """
        if not isinstance(location, str):
            raise ValueError('location must be a string')
        self._saver.save(self._sess, '%s/model.ckpt' % (location))
        with open('%s/model.json' % (location), 'w') as fopen:
            json.dump(
                {
                    'dictionary': self.dictionary,
                    'maxlen': self.maxlen,
                    'keys': self.keys,
                    'vectorized': self.vectorized.tolist(),
                    'embedding_size': self._embedding_size,
                },
                fopen,
            )

    def get_similarity(self, string, anchor = 0.5):
        """
        Return similar topics / influencers.

        Parameters
        ----------
        string: str
        anchor: float, (default=0.5)
            baseline similarity.

        Returns
        -------
        results: list of strings
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        if not isinstance(anchor, float):
            raise ValueError('anchor must be a float')
        if not (anchor > 0 and anchor < 1):
            raise ValueError('anchor must be bigger than 0, less than 1')
        original_string = simple_textcleaning(string)
        string = ' '.join(set(original_string.split()))
        encoded = self._sess.run(
            self._model.get_thought,
            feed_dict = {
                self._model.INPUT: batch_sequence(
                    [string], self.dictionary, maxlen = self.maxlen
                )
            },
        )
        where = np.where(
            cosine_similarity(self.vectorized, encoded)[:, 0] > anchor
        )[0]
        results = [self.keys[i].lower() for i in where]
        return list(set(results))


class _FAST_SIMILARITY:
    def __init__(self, vectorizer, vectorized, keys):
        self.vectorizer = vectorizer
        self.vectorized = vectorized
        self.keys = keys

    def get_similarity(self, string, anchor = 0.1):
        """
        Return similar topics / influencers.

        Parameters
        ----------
        string: str
        anchor: float, (default=0.5)
            baseline similarity.

        Returns
        -------
        results: list of strings
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        if not isinstance(anchor, float):
            raise ValueError('anchor must be a float')
        if not (anchor > 0 and anchor < 1):
            raise ValueError('anchor must be bigger than 0, less than 1')
        original_string = simple_textcleaning(string)
        string = ' '.join(set(original_string.split()))
        where = np.where(
            cosine_similarity(
                self.vectorized, self.vectorizer.transform([string])
            )[:, 0]
            > anchor
        )[0]
        results = [self.keys[i].lower() for i in where]
        return list(set(results))


class _FUZZY:
    def __init__(self, dictionary):
        self._dictionary = dictionary

    def get_similarity(self, string, fuzzy_ratio = 90):

        if not isinstance(string, str):
            raise ValueError('input must be a string')
        if not isinstance(fuzzy_ratio, int):
            raise ValueError('fuzzy_ratio must be an integer')
        if not (fuzzy_ratio > 0 and fuzzy_ratio < 100):
            raise ValueError(
                'fuzzy_ratio must be bigger than 0 and less than 100'
            )
        string = string.lower()
        topics = []
        for key, vals in self._dictionary.items():
            for v in vals:
                if fuzz.token_set_ratio(v, string) >= fuzzy_ratio:
                    topics.append(key.lower())
                    break
        return list(set(topics))


def fuzzy(dictionary):
    """
    Return similar topics.

    Parameters
    ----------
    dictionary: dict
        format {'left':['right']}

    Returns
    -------
    results: _FUZZY object
    """
    if not isinstance(dictionary, dict):
        raise ValueError('dictionary must be a dictionary')
    if not isinstance(list(dictionary.keys())[0], str):
        raise ValueError('keys dictionary must be a string')
    if not isinstance(list(dictionary.values())[0], list):
        raise ValueError('values dictionary must be a list')
    if not isinstance(list(dictionary.values())[0][0], str):
        raise ValueError('elements of values dictionary must be a string')
    return _FUZZY(dictionary)


def is_location(string, fuzzy_ratio = 90, location = _location):
    """
    check whether a string is a location, default is malaysia location.

    Parameters
    ----------
    string: str
    fuzzy_ratio: int, (default=90)
        ratio of similar characters by positions, if 90, means 90%
    location: list, (default=_location)
        list of locations

    Returns
    -------
    boolean: bool
    """
    if not isinstance(string, str):
        raise ValueError('input must be a string')
    if not isinstance(fuzzy_ratio, int):
        raise ValueError('fuzzy_ratio must be an integer')
    if not (fuzzy_ratio > 0 and fuzzy_ratio < 100):
        raise ValueError('fuzzy_ratio must be bigger than 0 and less than 100')
    for loc in location:
        if fuzz.token_set_ratio(loc.lower(), string) >= fuzzy_ratio:
            return True
    return False


def fuzzy_location(string, fuzzy_ratio = 90):
    """
    Return malaysia locations splitted by dun, parlimen and state from a string.

    Parameters
    ----------
    string: str
    fuzzy_ratio: int, (default=90)
        ratio of similar characters by positions, if 90, means 90%

    Returns
    -------
    results: list of strings
    """
    if not isinstance(string, str):
        raise ValueError('input must be a string')
    if not isinstance(fuzzy_ratio, int):
        raise ValueError('fuzzy_ratio must be an integer')
    if not (fuzzy_ratio > 0 and fuzzy_ratio < 100):
        raise ValueError('fuzzy_ratio must be bigger than 0 and less than 100')
    negeri_list = list(
        set(
            [
                i
                for i in _negeri
                if fuzz.token_set_ratio(i, string) >= fuzzy_ratio
            ]
        )
    )
    parlimen_list = list(
        set(
            [
                i
                for i in _parlimen
                if fuzz.token_set_ratio(i, string) >= fuzzy_ratio
            ]
        )
    )
    dun_list = list(
        set([i for i in _dun if fuzz.token_set_ratio(i, string) >= fuzzy_ratio])
    )
    return {'negeri': negeri_list, 'parlimen': parlimen_list, 'dun': dun_list}


def _generate_topics(dictionary):
    texts = [' '.join(words) for _, words in dictionary.items()]
    keys = [key for key, _ in dictionary.items()]
    texts = [' '.join(list(set(text.split()))) for text in texts]
    output = []
    for text in texts:
        output.append(
            ' '.join([word for word in text.split() if word not in STOPWORDS])
        )
    return output, keys


def deep_siamese(
    dictionary,
    epoch = 5,
    batch_size = 32,
    embedding_size = 256,
    output_size = 100,
    maxlen = 100,
    ngram = (1, 4),
    num_layers = 1,
    **kwargs
):
    """
    Train a deep siamese network for text similarity

    Parameters
    ----------
    dictionary: dict
        format {'left':['right']}
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    output_size: int, (default=100)
        encoder output size, bigger means more vector definition
    maxlen: int, (default=100)
        max length of a string to be train
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus
    num_layers: int, (default=100)
        number of bidirectional rnn layers

    Returns
    -------
    _DEEP_SIAMESE_SIMILARITY: malaya.similarity._DEEP_SIAMESE_SIMILARITY class
    """
    if not isinstance(dictionary, dict):
        raise ValueError('dictionary must be a dictionary')
    if not isinstance(list(dictionary.keys())[0], str):
        raise ValueError('keys dictionary must be a string')
    if not isinstance(list(dictionary.values())[0], list):
        raise ValueError('values dictionary must be a list')
    if not isinstance(list(dictionary.values())[0][0], str):
        raise ValueError('elements of values dictionary must be a string')
    if not isinstance(epoch, int):
        raise ValueError('epoch must be an integer')
    if not isinstance(batch_size, int):
        raise ValueError('batch_size must be an integer')
    if not isinstance(embedding_size, int):
        raise ValueError('embedding_size must be an integer')
    if not isinstance(output_size, int):
        raise ValueError('output_size must be an integer')
    if not isinstance(maxlen, int):
        raise ValueError('maxlen must be an integer')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not isinstance(num_layers, int):
        raise ValueError('num_layers must be an integer')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    output, keys = _generate_topics(dictionary)
    batch_x_left, batch_x_right, batch_y = [], [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_x_right.extend([keys[i]] * len(augmentation))
        batch_x_left.extend(augmentation)
        batch_y.extend([1] * len(augmentation))
    neg_batch_x_left, neg_batch_x_right = [], []
    for i in range(len(batch_x_left)):
        random_num = random.randint(0, len(batch_x_left) - 1)
        while random_num == i:
            random_num = random.randint(0, len(batch_x_left) - 1)
        neg_batch_x_right.append(batch_x_right[random_num])
        neg_batch_x_left.append(batch_x_left[i])
        batch_y.append(0)
    batch_x_left.extend(neg_batch_x_left)
    batch_x_right.extend(neg_batch_x_right)
    batch_x_left, batch_x_right, batch_y = shuffle(
        batch_x_left, batch_x_right, batch_y
    )
    sess, model, dictionary, saver, dropout = siamese_train(
        batch_x_left,
        batch_x_right,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        output_size = output_size,
        maxlen = maxlen,
        num_layers = num_layers,
        **kwargs
    )
    return _DEEP_SIAMESE_SIMILARITY(
        sess,
        model,
        keys,
        dictionary,
        maxlen,
        saver,
        embedding_size,
        num_layers,
        output_size,
        dropout,
    )


def load_siamese(location):
    if not os.path.isfile('%s/model.json' % (location)) or not os.path.isfile(
        '%s/model.ckpt.meta' % (location)
    ):
        raise Exception(
            'siamese text similarity model at %s/ is not available, please train and save the model'
            % (location)
        )
    with open('%s/model.json' % (location)) as fopen:
        json_ = json.load(fopen)
    sess, model, saver = load_deep_siamese(location, json_)
    return _DEEP_SIAMESE_SIMILARITY(
        sess,
        model,
        json_['keys'],
        json_['dictionary'],
        json_['maxlen'],
        saver,
        json_['embedding_size'],
        json_['num_layers'],
        json_['output_size'],
        json_['dropout'],
    )


def deep_skipthought(
    dictionary,
    epoch = 5,
    batch_size = 16,
    embedding_size = 256,
    maxlen = 100,
    ngram = (1, 4),
):
    """
    Train a deep skip-thought network for text similarity

    Parameters
    ----------
    dictionary: dict
        format {'left':['right']}
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    maxlen: int, (default=100)
        max length of a string to be train
    ngram: tuple, (default=(1,4))
        n-grams size to train a corpus

    Returns
    -------
    _DEEP_SIMILARITY: malaya.similarity._DEEP_SIMILARITY class
    """
    if not isinstance(dictionary, dict):
        raise ValueError('dictionary must be a dictionary')
    if not isinstance(list(dictionary.keys())[0], str):
        raise ValueError('keys dictionary must be a string')
    if not isinstance(list(dictionary.values())[0], list):
        raise ValueError('values dictionary must be a list')
    if not isinstance(list(dictionary.values())[0][0], str):
        raise ValueError('elements of values dictionary must be a string')
    if not isinstance(epoch, int):
        raise ValueError('epoch must be an integer')
    if not isinstance(batch_size, int):
        raise ValueError('batch_size must be an integer')
    if not isinstance(embedding_size, int):
        raise ValueError('embedding_size must be an integer')
    if not isinstance(maxlen, int):
        raise ValueError('maxlen must be an integer')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    output, keys = _generate_topics(dictionary)
    batch_x, batch_y = [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_y.extend([keys[i]] * len(augmentation))
        batch_x.extend(augmentation)
    batch_x, batch_y = shuffle(batch_x, batch_y)
    sess, model, dictionary, saver = skip_train(
        batch_x,
        batch_y,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        maxlen = maxlen,
    )
    vectorized = sess.run(
        model.get_thought,
        feed_dict = {
            model.INPUT: batch_sequence(output, dictionary, maxlen = maxlen)
        },
    )
    return _DEEP_SIMILARITY(
        sess, model, vectorized, keys, dictionary, maxlen, saver, embedding_size
    )


def load_skipthought(location):
    if not os.path.isfile('%s/model.json' % (location)) or not os.path.isfile(
        '%s/model.ckpt.meta' % (location)
    ):
        raise Exception(
            'skipthought text similarity model at %s/ is not available, please train and save the model'
        )
    with open('%s/model.json' % (location)) as fopen:
        json_ = json.load(fopen)
    sess, model, saver = load_deep_skipthought(location, json_)
    return _DEEP_SIMILARITY(
        sess,
        model,
        np.array(json_['vectorized']),
        json_['keys'],
        json_['dictionary'],
        json_['maxlen'],
        saver,
        json_['embedding_size'],
    )


def bow(dictionary, vectorizer = 'tfidf', ngram = (3, 10)):
    """
    Train a bow for text similarity

    Parameters
    ----------
    dictionary: dict
        format {'left':['right']}
    vectorizer: str, (default='tfidf')
        vectorization technique for a corpus
    ngram: tuple, (default=(3,10))
        n-grams size to train a corpus

    Returns
    -------
    _FAST_SIMILARITY: malaya.similarity._FAST_SIMILARITY class
    """
    if not isinstance(dictionary, dict):
        raise ValueError('dictionary must be a dictionary')
    if not isinstance(list(dictionary.keys())[0], str):
        raise ValueError('keys dictionary must be a string')
    if not isinstance(list(dictionary.values())[0], list):
        raise ValueError('values dictionary must be a list')
    if not isinstance(list(dictionary.values())[0][0], str):
        raise ValueError('elements of values dictionary must be a string')
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if 'tfidf' in vectorizer.lower():
        char_vectorizer = TfidfVectorizer(
            sublinear_tf = True,
            strip_accents = 'unicode',
            analyzer = 'char',
            ngram_range = ngram,
        )
    elif 'count' in vectorizer.lower():
        char_vectorizer = CountVectorizer(
            strip_accents = 'unicode', analyzer = 'char', ngram_range = ngram
        )
    elif 'skip-gram' in vectorizer.lower():
        char_vectorizer = SkipGramVectorizer(
            strip_accents = 'unicode', analyzer = 'char', ngram_range = ngram
        )
    else:
        raise Exception('model not supported')
    output, keys = _generate_topics(dictionary)
    vectorizer = char_vectorizer.fit(output)
    vectorized = vectorizer.transform(output)
    return _FAST_SIMILARITY(vectorizer, vectorized, keys)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import os
import numpy as np
from fuzzywuzzy import fuzz
from sklearn.manifold import TSNE
from scipy.spatial.distance import cdist
from sklearn.neighbors import NearestNeighbors
import tensorflow as tf
from . import home
from ._utils._utils import download_file
from ._models import _word2vec
from .texts._text_functions import simple_textcleaning


def load_wiki():
    """
    Return malaya pretrained wikipedia word2vec size 256

    Returns
    -------
    dictionary: dictionary of dictionary, reverse dictionary and vectors
    """
    if not os.path.isfile('%s/word2vec-wiki/word2vec.p' % (home)):
        print('downloading word2vec-wiki embedded')
        download_file(
            'v13/word2vec/word2vec-wiki-nce-256.p',
            '%s/word2vec-wiki/word2vec.p' % (home),
        )
    with open('%s/word2vec-wiki/word2vec.p' % (home), 'rb') as fopen:
        return pickle.load(fopen)


def load_news(size = 256):
    """
    Return malaya pretrained news word2vec

    Parameters
    ----------
    size: int, (default=256)

    Returns
    -------
    dictionary: dictionary of dictionary, reverse dictionary and vectors
    """
    if not isinstance(size, int):
        raise ValueError('input must be an integer')
    if size not in [32, 64, 128, 256, 512]:
        raise ValueError(
            'size word2vec not supported, only supports [32, 64, 128, 256, 512]'
        )
    if not os.path.isfile('%s/word2vec-%d/word2vec.p' % (home, size)):
        print('downloading word2vec-%d embedded' % (size))
        download_file(
            'v7/word2vec/word2vec-%d.p' % (size),
            '%s/word2vec-%d/word2vec.p' % (home, size),
        )
    with open('%s/word2vec-%d/word2vec.p' % (home, size), 'rb') as fopen:
        return pickle.load(fopen)


def train(
    corpus,
    vocab_size = None,
    batch_size = 32,
    embedding_size = 256,
    hidden_size = 256,
    negative_samples_ratio = 0.5,
    learning_rate = 0.01,
    embedding_noise = 0.1,
    hidden_noise = 0.3,
    momentum = 0.9,
    epoch = 10,
    optimizer = 'momentum',
    text_cleaning = simple_textcleaning,
):
    """
    Train a word2vec for custom corpus

    Parameters
    ----------
    corpus: list
        list of strings
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    hidden_size: int, (default=256)
        vector size representation for hidden layer
    negative_samples_ratio: float, (default=0.5)
        negative samples ratio proportional to batch_size
    learning_rate: float, (default=0.01)
        learning rate for word2vec
    momentum: float, (default=0.9)
        momentum rate for optimizer=momentum
    epoch: int, (default=5)
        iteration numbers
    optimizer: str, (default='momentum')
        optimizer supported, ['gradientdescent', 'rmsprop', 'momentum', 'adagrad', 'adam']
    text_cleaning: function, (default=simple_textcleaning)
        function to clean the corpus

    Returns
    -------
    dictionary: dictionary of dictionary, reverse dictionary and vectors
    """
    if not isinstance(corpus, str) and not isinstance(corpus, list):
        raise ValueError('corpus must be a string or a list of string')
    if not vocab_size is None and not isinstance(vocab_size, int):
        raise ValueError('vocab_size must be a None or an integer')
    if not isinstance(batch_size, int):
        raise ValueError('batch_size must be an integer')
    if not isinstance(embedding_size, int):
        raise ValueError('embedding_size must be an integer')
    if not isinstance(hidden_size, int):
        raise ValueError('hidden_size must be an integer')
    if not isinstance(epoch, int):
        raise ValueError('epoch must be an integer')
    if not isinstance(negative_samples_ratio, float):
        raise ValueError('negative_samples_ratio must be a float')
    if not isinstance(momentum, float):
        raise ValueError('momentum must be a float')
    if not isinstance(embedding_noise, float):
        raise ValueError('embedding_noise must be a float')
    if not isinstance(hidden_noise, float):
        raise ValueError('hidden_noise must be a float')
    if not isinstance(learning_rate, float) or isinstance(learning_rate, int):
        raise ValueError('learning_rate must be a float or an integer')
    if not isinstance(optimizer, str):
        raise ValueError('optimizer must be a string')
    if not batch_size > 0:
        raise ValueError('batch_size must bigger than 0')
    if not epoch > 0:
        raise ValueError('epoch must bigger than 0')
    if not embedding_size > 0:
        raise ValueError('embedding_size must bigger than 0')
    if not hidden_size > 0:
        raise ValueError('hidden_size must bigger than 0')
    if not (negative_samples_ratio > 0 and negative_samples_ratio <= 1):
        raise ValueError(
            'negative_samples_ratio must bigger than 0 and less than or equal 1'
        )
    if not (embedding_noise > 0 and embedding_noise <= 1):
        raise ValueError(
            'embedding_noise must bigger than 0 and less than or equal 1'
        )
    if not (hidden_noise > 0 and hidden_noise <= 1):
        raise ValueError(
            'hidden_noise must bigger than 0 and less than or equal 1'
        )
    optimizer = optimizer.lower()
    if optimizer not in [
        'gradientdescent',
        'adam',
        'adagrad',
        'momentum',
        'rmsprop',
    ]:
        raise Exception(
            "Optimizer not supported, only supports ['gradientdescent', 'rmsprop', 'momentum', 'adagrad', 'adam']"
        )
    from sklearn.model_selection import train_test_split

    if isinstance(corpus, list):
        corpus = ' '.join(corpus)
    if text_cleaning:
        corpus = text_cleaning(corpus)
    corpus = ' '.join(corpus.split('\n'))
    corpus = list(filter(None, corpus.split()))
    if vocab_size is None:
        vocab_size = len(set(corpus)) + 5
    word_array, dictionary, rev_dictionary, num_lines, num_words = _word2vec.build_word_array(
        corpus, vocab_size
    )
    X, Y = _word2vec.build_training_set(word_array)
    graph_params = {
        'batch_size': batch_size,
        'vocab_size': np.max(X) + 1,
        'embed_size': embedding_size,
        'hid_size': hidden_size,
        'neg_samples': int(batch_size * negative_samples_ratio),
        'learn_rate': learning_rate,
        'momentum': momentum,
        'embed_noise': embedding_noise,
        'hid_noise': hidden_noise,
        'epoch': epoch,
        'optimizer': optimizer,
    }
    _, test_X, _, test_Y = train_test_split(X, Y, test_size = 0.1)
    model = _word2vec.Model(graph_params)
    print(
        'model built, vocab size %d, document length %d'
        % (np.max(X) + 1, len(word_array))
    )
    embed_weights, nce_weights = model.train(
        X, Y, test_X, test_Y, graph_params['epoch'], graph_params['batch_size']
    )
    return embed_weights, nce_weights, dictionary


class _Calculator:
    def __init__(self, tokens):
        self._tokens = tokens
        self._current = tokens[0]

    def exp(self):
        result = self.term()
        while self._current in ('+', '-'):
            if self._current == '+':
                self.next()
                result += self.term()
            if self._current == '-':
                self.next()
                result -= self.term()
        return result

    def factor(self):
        result = None
        if self._current[0].isdigit() or self._current[-1].isdigit():
            result = np.array([float(i) for i in self._current.split(',')])
            self.next()
        elif self._current is '(':
            self.next()
            result = self.exp()
            self.next()
        return result

    def next(self):
        self._tokens = self._tokens[1:]
        self._current = self._tokens[0] if len(self._tokens) > 0 else None

    def term(self):
        result = self.factor()
        while self._current in ('*', '/'):
            if self._current == '*':
                self.next()
                result *= self.term()
            if self._current == '/':
                self.next()
                result /= self.term()
        return result


class word2vec:
    def __init__(self, embed_matrix, dictionary, no_cores = None):
        self._embed_matrix = embed_matrix
        self._dictionary = dictionary
        self._reverse_dictionary = {v: k for k, v in dictionary.items()}
        self.words = list(dictionary.keys())
        _graph = tf.Graph()
        with _graph.as_default():
            self._embedding = tf.placeholder(
                tf.float32, self._embed_matrix.shape
            )
            self._x = tf.placeholder(
                tf.float32, [None, self._embed_matrix.shape[1]]
            )
            normed_embedding = tf.nn.l2_normalize(self._embedding, axis = 1)
            normed_array = tf.nn.l2_normalize(self._x, axis = 1)
            self._cosine_similarity = tf.matmul(
                normed_array, tf.transpose(normed_embedding, [1, 0])
            )
            if no_cores:
                config = tf.ConfigProto()
                config.intra_op_parallelism_threads = no_cores
                config.intra_op_parallelism_threads = no_cores
                self._sess = tf.InteractiveSession(config = config)
            else:
                self._sess = tf.InteractiveSession()

    def get_vector_by_name(self, word):
        """
        get vector based on name

        Parameters
        ----------
        word: str

        Returns
        -------
        vector: numpy
        """
        if word not in self._dictionary:
            arr = np.array([fuzz.ratio(word, k) for k in self.words])
            idx = (-arr).argsort()[:5]
            strings = ', '.join([self.words[i] for i in idx])
            raise Exception(
                'input not found in dictionary, here top-5 nearest words [%s]'
                % (strings)
            )
        return np.ravel(self._embed_matrix[self._dictionary[word], :])

    def tree_plot(
        self,
        labels,
        notebook_mode = False,
        figsize = (7, 7),
        annotate = True,
        figname = 'fig.png',
    ):
        """
        plot a tree plot based on output from calculator / n_closest / analogy

        Parameters
        ----------
        labels : list
            output from calculator / n_closest / analogy
        notebook_mode : bool
            if True, it will render plt.show, else plt.savefig
        figsize : tuple, (default=(7, 7))
            figure size for plot
        figname : str, (default='fig.png')

        Returns
        -------
        list_dictionaries: list of results
        """
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns

            sns.set()
        except:
            raise Exception(
                'matplotlib and seaborn not installed. Please install it and try again.'
            )
        idx = [
            self.words.index(e[0] if isinstance(e, list) else e) for e in labels
        ]
        embed = self._embed_matrix[idx]
        embed = embed.dot(embed.T)
        embed = (embed - embed.min()) / (embed.max() - embed.min())
        labelled = []
        for label in labels:
            label = (
                '%s, %.3f' % (label[0], label[1])
                if isinstance(label, list)
                else label
            )
            labelled.append(label)
        plt.figure(figsize = figsize)
        g = sns.clustermap(
            embed,
            cmap = 'Blues',
            xticklabels = labelled,
            yticklabels = labelled,
            annot = annotate,
        )
        if notebook_mode:
            plt.show()
        else:
            plt.savefig(figname, bbox_inches = 'tight')

    def scatter_plot(
        self,
        labels,
        centre = None,
        notebook_mode = False,
        figsize = (7, 7),
        figname = 'fig.png',
        plus_minus = 25,
        handoff = 5e-5,
    ):
        """
        plot a scatter plot based on output from calculator / n_closest / analogy

        Parameters
        ----------
        labels : list
            output from calculator / n_closest / analogy
        centre : str, (default=None)
            centre label, if a str, it will annotate in a red color
        notebook_mode : bool
            if True, it will render plt.show, else plt.savefig
        figsize : tuple, (default=(7, 7))
            figure size for plot
        figname : str, (default='fig.png')

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(labels, list):
            raise ValueError('input must be a list')
        if not isinstance(notebook_mode, bool):
            raise ValueError('notebook_mode must be a boolean')
        if not isinstance(figsize, tuple):
            raise ValueError('figsize must be a tuple')
        if not isinstance(figname, str):
            raise ValueError('figname must be a string')
        if not isinstance(plus_minus, int):
            raise ValueError('plus_minus must be an integer')
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns

            sns.set()
        except:
            raise Exception(
                'matplotlib and seaborn not installed. Please install it and try again.'
            )
        idx = [
            self.words.index(e[0] if isinstance(e, list) else e) for e in labels
        ]
        if centre:
            if not isinstance(centre, str):
                raise ValueError('centre must be a string')
            idx.append(self.words.index(centre))
        cp_idx = idx[:]
        for i in idx:
            cp_idx.extend(np.arange(i - plus_minus, i).tolist())
            cp_idx.extend(np.arange(i, i + plus_minus).tolist())
        tsne = TSNE(n_components = 2, random_state = 0).fit_transform(
            self._embed_matrix[cp_idx]
        )
        plt.figure(figsize = figsize)
        plt.scatter(tsne[:, 0], tsne[:, 1])
        for label, x, y in zip(
            labels, tsne[: len(labels), 0], tsne[: len(labels), 1]
        ):
            label = (
                '%s, %.3f' % (label[0], label[1])
                if isinstance(label, list)
                else label
            )
            plt.annotate(
                label,
                xy = (x, y),
                xytext = (0, 0),
                textcoords = 'offset points',
            )
        if centre:
            plt.annotate(
                centre,
                xy = (tsne[len(labels), 0], tsne[len(labels), 1]),
                xytext = (0, 0),
                textcoords = 'offset points',
                color = 'red',
            )
        plt.xlim(
            tsne[: len(idx), 0].min() + handoff,
            tsne[: len(idx), 0].max() + handoff,
        )
        plt.ylim(
            tsne[: len(idx), 1].min() + handoff,
            tsne[: len(idx), 1].max() + handoff,
        )
        plt.xticks([])
        plt.yticks([])
        if notebook_mode:
            plt.show()
        else:
            plt.savefig(figname, bbox_inches = 'tight')

    def _calculate(self, equation):
        tokens, temp = [], ''
        for char in equation:
            if char == ' ':
                continue
            if char not in '()*+-':
                temp += char
            else:
                if len(temp):
                    row = self._dictionary[
                        self.words[
                            np.argmax([fuzz.ratio(temp, k) for k in self.words])
                        ]
                    ]
                    tokens.append(
                        ','.join(
                            self._embed_matrix[row, :].astype('str').tolist()
                        )
                    )
                    temp = ''
                tokens.append(char)
        if len(temp):
            row = self._dictionary[
                self.words[np.argmax([fuzz.ratio(temp, k) for k in self.words])]
            ]
            tokens.append(
                ','.join(self._embed_matrix[row, :].astype('str').tolist())
            )
        return _Calculator(tokens).exp()

    def _batch_process(self, batch, num_closest = 5, return_similarity = True):
        top_k = tf.nn.top_k(self._cosine_similarity, k = num_closest)
        results = self._sess.run(
            top_k,
            feed_dict = {self._x: batch, self._embedding: self._embed_matrix},
        )
        indices = results.indices
        values = results.values
        words = []
        if not return_similarity:
            for result in indices:
                words.append([self._reverse_dictionary[i] for i in result])
        else:
            for no in range(len(results)):
                words.append(
                    [
                        (
                            self._reverse_dictionary[indices[no, i]],
                            values[no, i],
                        )
                        for i in range(len(indices[no]))
                    ]
                )
        return words

    def batch_calculator(
        self, equations, num_closest = 5, return_similarity = False
    ):
        """
        batch calculator parser for word2vec using tensorflow

        Parameters
        ----------
        equations: list of str
            Eg, '[(mahathir + najib) - rosmah]'
        num_closest: int, (default=5)
            number of words closest to the result

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(equations, list):
            raise ValueError('equations must be a list of string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        batches = np.array([self._calculate(eq) for eq in equations])
        return self._batch_process(
            batches,
            num_closest = num_closest,
            return_similarity = return_similarity,
        )

    def calculator(
        self,
        equation,
        num_closest = 5,
        metric = 'cosine',
        return_similarity = True,
    ):
        """
        calculator parser for word2vec

        Parameters
        ----------
        equation: str
            Eg, '(mahathir + najib) - rosmah'
        num_closest: int, (default=5)
            number of words closest to the result
        metric: str, (default='cosine')
            vector distance algorithm
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(equation, str):
            raise ValueError('input must be a string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(metric, str):
            raise ValueError('metric must be a string')
        if not isinstance(return_similarity, bool):
            raise ValueError('num_closest must be a boolean')
        calculated = self._calculate(equation)
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(calculated.reshape((1, -1)))
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            closest_indices = self.closest_row_indices(
                calculated, num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            return word_list

    def batch_n_closest(
        self, words, num_closest = 5, return_similarity = False, soft = True
    ):
        """
        find nearest words based on a batch of words using Tensorflow

        Parameters
        ----------
        words: list
            Eg, ['najib','anwar']
        num_closest: int, (default=5)
            number of words closest to the result
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance
        soft: bool, (default=True)
            if True, a word not in the dictionary will be replaced with nearest fuzzywuzzy ratio.
            if False, it will throw an exception if a word not in the dictionary

        Returns
        -------
        word_list: list of nearest words
        """
        if soft:
            for i in range(len(words)):
                if words[i] not in self.words:
                    words[i] = self.words[
                        np.argmax([fuzz.ratio(words[i], k) for k in self.words])
                    ]
        else:
            for i in range(len(words)):
                if words[i] not in self.words:
                    raise Exception(
                        '%s not in dictionary, please use another word or set `soft` = True'
                        % (words[i])
                    )
        batches = np.array([self.get_vector_by_name(w) for w in words])
        return self._batch_process(
            batches,
            num_closest = num_closest,
            return_similarity = return_similarity,
        )

    def n_closest(
        self, word, num_closest = 5, metric = 'cosine', return_similarity = True
    ):
        """
        find nearest words based on a word

        Parameters
        ----------
        word: str
            Eg, 'najib'
        num_closest: int, (default=5)
            number of words closest to the result
        metric: str, (default='cosine')
            vector distance algorithm
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(word, str):
            raise ValueError('input must be a string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(metric, str):
            raise ValueError('metric must be a string')
        if not isinstance(return_similarity, bool):
            raise ValueError('num_closest must be a boolean')
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(
                self.get_vector_by_name(word).reshape((1, -1))
            )
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            wv = self.get_vector_by_name(word)
            closest_indices = self.closest_row_indices(
                wv, num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            if word in word_list:
                word_list.remove(word)
            return word_list

    def closest_row_indices(self, wv, num, metric):
        dist_array = np.ravel(
            cdist(self._embed_matrix, wv.reshape((1, -1)), metric = metric)
        )
        sorted_indices = np.argsort(dist_array)
        return sorted_indices[:num]

    def analogy(self, a, b, c, num = 1, metric = 'cosine'):
        """
        analogy calculation, vb - va + vc

        Parameters
        ----------
        a: str
        b: str
        c: str
        num: int, (default=1)
        metric: str, (default='cosine')
            vector distance algorithm

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(a, str):
            raise ValueError('a must be a string')
        if not isinstance(b, str):
            raise ValueError('b must be a string')
        if not isinstance(c, str):
            raise ValueError('c must be a string')
        if a not in self._dictionary:
            raise Exception('a not in dictinary')
        if b not in self._dictionary:
            raise Exception('b not in dictinary')
        if c not in self._dictionary:
            raise Exception('c not in dictinary')
        va = self.get_vector_by_name(a)
        vb = self.get_vector_by_name(b)
        vc = self.get_vector_by_name(c)
        vd = vb - va + vc
        closest_indices = self.closest_row_indices(vd, num, metric)
        d_word_list = []
        for i in closest_indices:
            d_word_list.append(self._reverse_dictionary[i])
        return d_word_list

    def project_2d(self, start, end):
        """
        project word2vec into 2d dimension

        Parameters
        ----------
        start: int
        end: int

        Returns
        -------
        tsne decomposition: numpy
        """
        if not isinstance(start, int):
            raise ValueError('start must be an integer')
        if not isinstance(end, int):
            raise ValueError('end must be an integer')
        tsne = TSNE(n_components = 2)
        embed_2d = tsne.fit_transform(self._embed_matrix[start:end, :])
        word_list = []
        for i in range(start, end):
            word_list.append(self._reverse_dictionary[i])
        return embed_2d, word_list



from sklearn.metrics.pairwise import cosine_similarity
from sklearn.cluster import KMeans
from sklearn.manifold import MDS
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from .texts.vectorizer import SkipGramVectorizer
from .stem import sastrawi
from .texts._text_functions import simple_textcleaning, STOPWORDS
from scipy.cluster.hierarchy import ward, dendrogram
import numpy as np
import random

_accepted_pos = [
    'ADJ',
    'ADP',
    'ADV',
    'ADX',
    'CCONJ',
    'DET',
    'NOUN',
    'NUM',
    'PART',
    'PRON',
    'PROPN',
    'SCONJ',
    'SYM',
    'VERB',
    'X',
]
_accepted_entities = [
    'OTHER',
    'law',
    'location',
    'organization',
    'person',
    'quantity',
    'time',
    'event',
]


def cluster_words(list_words):
    """
    cluster similar words based on structure, eg, ['mahathir mohamad', 'mahathir'] = ['mahathir mohamad']

    Parameters
    ----------
    list_words : list of str

    Returns
    -------
    string: list of clustered words
    """
    if not isinstance(list_words, list):
        raise ValueError('list_words must be a list')
    if not isinstance(list_words[0], str):
        raise ValueError('list_words must be a list of strings')

    dict_words = {}
    for word in list_words:
        found = False
        for key in dict_words.keys():
            if word in key or any(
                [word in inside for inside in dict_words[key]]
            ):
                dict_words[key].append(word)
                found = True
            if key in word:
                dict_words[key].append(word)
        if not found:
            dict_words[word] = [word]
    results = []
    for key, words in dict_words.items():
        results.append(max(list(set([key] + words)), key = len))
    return list(set(results))


def cluster_pos(result):
    """
    cluster similar POS

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    if not isinstance(result, list):
        raise ValueError('result must be a list')
    if not isinstance(result[0], tuple):
        raise ValueError('result must be a list of tuple')
    if not all([i[1] in _accepted_pos for i in result]):
        raise ValueError(
            'elements of result must be a subset or equal of supported POS, please run malaya.describe_pos() to get supported POS'
        )

    output = {
        'ADJ': [],
        'ADP': [],
        'ADV': [],
        'ADX': [],
        'CCONJ': [],
        'DET': [],
        'NOUN': [],
        'NUM': [],
        'PART': [],
        'PRON': [],
        'PROPN': [],
        'SCONJ': [],
        'SYM': [],
        'VERB': [],
        'X': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    return output


def cluster_entities(result):
    """
    cluster similar Entities

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    if not isinstance(result, list):
        raise ValueError('result must be a list')
    if not isinstance(result[0], tuple):
        raise ValueError('result must be a list of tuple')
    if not all([i[1] in _accepted_entities for i in result]):
        raise ValueError(
            'elements of result must be a subset or equal of supported POS, please run malaya.describe_pos() to get supported POS'
        )

    output = {
        'OTHER': [],
        'law': [],
        'location': [],
        'organization': [],
        'person': [],
        'quantity': [],
        'time': [],
        'event': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    return output


def cluster_scatter(
    corpus,
    titles = None,
    colors = None,
    stemming = True,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    cleaning = simple_textcleaning,
    vectorizer = 'bow',
    stop_words = STOPWORDS,
    num_clusters = 5,
    clustering = KMeans,
    decomposition = MDS,
    figsize = (17, 9),
):
    """
    corpus: list
    titles: list
        list of titles, length must same with corpus
    colors: list
        list of colors, length must same with num_clusters
    num_clusters: int, (default=5)
        size of unsupervised clusters.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word
        * ``'tfidf'`` - Term frequency inverse Document Frequency
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams
    """
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(titles, list) and titles is not None:
        raise ValueError('titles must be a list or None')
    if not isinstance(colors, list) and colors is not None:
        raise ValueError('colors must be a list or None')
    if titles:
        if len(titles) != len(corpus):
            raise ValueError('length of titles must be same with corpus')
    if colors:
        if len(colors) != num_clusters:
            raise ValueError(
                'size of colors must be same with number of clusters'
            )
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')
    vectorizer = vectorizer.lower()
    if not vectorizer in ['tfidf', 'bow', 'skip-gram']:
        raise ValueError("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not (isinstance(max_df, int) or isinstance(max_df, float)):
        raise ValueError('max_df must be an integer or a float')
    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df < 1 and max_df > 0):
        raise ValueError('max_df must be bigger than 0, less than 1')
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns

        sns.set()
    except:
        raise Exception(
            'matplotlib and seaborn not installed. Please install it and try again.'
        )

    tf_vectorizer = Vectorizer(
        ngram_range = ngram,
        min_df = min_df,
        max_df = max_df,
        stop_words = stop_words,
    )
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )
    tf_vectorizer.fit(text_clean)
    transformed_text_clean = tf_vectorizer.transform(text_clean)
    features = tf_vectorizer.get_feature_names()
    km = clustering(n_clusters = num_clusters)
    dist = 1 - cosine_similarity(transformed_text_clean)
    km.fit(transformed_text_clean)
    clusters = km.labels_.tolist()
    if isinstance(decomposition, MDS):
        decomposed = decomposition(
            n_components = 2, dissimilarity = 'precomputed'
        )
    else:
        decomposed = decomposition(n_components = 2)
    pos = decomposed.fit_transform(dist)
    if not titles:
        titles = []
        for i in range(transformed_text_clean.shape[0]):
            indices = np.argsort(
                np.array(transformed_text_clean[i].todense())[0]
            )[::-1]
            titles.append(' '.join([features[i] for i in indices[: ngram[1]]]))
    if not colors:
        colors = sns.color_palette(n_colors = num_clusters)
    X, Y = pos[:, 0], pos[:, 1]
    plt.figure(figsize = figsize)
    for i in np.unique(clusters):
        plt.scatter(
            X[clusters == i],
            Y[clusters == i],
            color = colors[i],
            label = 'cluster %d' % (i),
        )
    for i in range(len(X)):
        plt.text(X[i], Y[i], titles[i], size = 8)
    plt.legend()
    plt.show()
    return {
        'X': X,
        'Y': Y,
        'labels': clusters,
        'vector': transformed_text_clean,
        'titles': titles,
    }


def cluster_dendogram(
    corpus,
    titles = None,
    stemming = True,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    cleaning = simple_textcleaning,
    vectorizer = 'bow',
    stop_words = STOPWORDS,
    random_samples = 0.3,
    figsize = (17, 9),
):
    """
    corpus: list
    titles: list
        list of titles, length must same with corpus
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word
        * ``'tfidf'`` - Term frequency inverse Document Frequency
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams
    """
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(titles, list) and titles is not None:
        raise ValueError('titles must be a list or None')
    if titles:
        if len(titles) != len(corpus):
            raise ValueError('length of titles must be same with corpus')
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')
    vectorizer = vectorizer.lower()
    if not vectorizer in ['tfidf', 'bow', 'skip-gram']:
        raise ValueError("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not (isinstance(max_df, int) or isinstance(max_df, float)):
        raise ValueError('max_df must be an integer or a float')
    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df < 1 and max_df > 0):
        raise ValueError('max_df must be bigger than 0, less than 1')
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns

        sns.set()
    except:
        raise Exception(
            'matplotlib and seaborn not installed. Please install it and try again.'
        )

    tf_vectorizer = Vectorizer(
        ngram_range = ngram,
        min_df = min_df,
        max_df = max_df,
        stop_words = stop_words,
    )
    corpus = random.sample(corpus, k = int(random_samples * len(corpus)))
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )
    tf_vectorizer.fit(text_clean)
    transformed_text_clean = tf_vectorizer.transform(text_clean)
    features = tf_vectorizer.get_feature_names()
    dist = 1 - cosine_similarity(transformed_text_clean)
    linkage_matrix = ward(dist)
    if not titles:
        titles = []
        for i in range(transformed_text_clean.shape[0]):
            indices = np.argsort(
                np.array(transformed_text_clean[i].todense())[0]
            )[::-1]
            titles.append(' '.join([features[i] for i in indices[: ngram[1]]]))
    plt.figure(figsize = figsize)
    ax = dendrogram(linkage_matrix, orientation = 'right', labels = titles)
    plt.tick_params(
        axis = 'x',
        which = 'both',
        bottom = 'off',
        top = 'off',
        labelbottom = 'off',
    )
    plt.tight_layout()
    plt.show()
    return {'linkage_matrix': linkage_matrix, 'titles': titles}


def cluster_graph(
    corpus,
    titles = None,
    colors = None,
    threshold = 0.3,
    stemming = True,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    cleaning = simple_textcleaning,
    vectorizer = 'bow',
    stop_words = STOPWORDS,
    num_clusters = 5,
    clustering = KMeans,
    figsize = (17, 9),
    with_labels = True,
):
    """
    corpus: list
    titles: list
        list of titles, length must same with corpus
    colors: list
        list of colors, length must same with num_clusters
    num_clusters: int, (default=5)
        size of unsupervised clusters.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus
    stop_words: list, (default=STOPWORDS)
        list of stop words to remove
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word
        * ``'tfidf'`` - Term frequency inverse Document Frequency
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams
    """
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(titles, list) and titles is not None:
        raise ValueError('titles must be a list or None')
    if not isinstance(colors, list) and colors is not None:
        raise ValueError('colors must be a list or None')
    if titles:
        if len(titles) != len(corpus):
            raise ValueError('length of titles must be same with corpus')
    if colors:
        if len(colors) != num_clusters:
            raise ValueError(
                'size of colors must be same with number of clusters'
            )
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')
    vectorizer = vectorizer.lower()
    if not vectorizer in ['tfidf', 'bow', 'skip-gram']:
        raise ValueError("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not (isinstance(max_df, int) or isinstance(max_df, float)):
        raise ValueError('max_df must be an integer or a float')
    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df < 1 and max_df > 0):
        raise ValueError('max_df must be bigger than 0, less than 1')
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns
        import networkx as nx
        import networkx.drawing.layout as nxlayout

        sns.set()
    except:
        raise Exception(
            'matplotlib, seaborn, networkx not installed. Please install it and try again.'
        )

    tf_vectorizer = Vectorizer(
        ngram_range = ngram,
        min_df = min_df,
        max_df = max_df,
        stop_words = stop_words,
    )
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )
    tf_vectorizer.fit(text_clean)
    DxT = tf_vectorizer.transform(text_clean)
    DxD = np.dot(DxT, DxT.T)
    km = clustering(n_clusters = num_clusters)
    km.fit(DxT)
    clusters = km.labels_.tolist()
    features = tf_vectorizer.get_feature_names()
    if not titles:
        titles = []
        for i in range(DxT.shape[0]):
            indices = np.argsort(np.array(DxT[i].todense())[0])[::-1]
            titles.append(' '.join([features[i] for i in indices[: ngram[1]]]))
    if not colors:
        colors = sns.color_palette(n_colors = num_clusters)
    G = nx.Graph()
    for i in range(DxT.shape[0]):
        G.add_node(i, text = titles[i], label = clusters[i])
    dense_DxD = DxD.toarray()
    len_dense = len(dense_DxD)
    for i in range(len_dense):
        for j in range(i + 1, len_dense):
            if dense_DxD[i, j] >= threshold:
                weight = dense_DxD[i, j]
                G.add_edge(i, j, weight = weight)
    for node, degree in list(dict(G.degree()).items()):
        if degree == 0:
            G.remove_node(node)
    node_colors, node_labels = [], {}
    for node in G:
        node_colors.append(colors[G.node[node]['label']])
        node_labels[node] = G.node[node]['text']
    pos = nxlayout.fruchterman_reingold_layout(
        G, k = 1.5 / np.sqrt(len(G.nodes()))
    )
    plt.figure(figsize = figsize)
    if with_labels:
        nx.draw(G, node_color = node_colors, pos = pos, labels = node_labels)
    else:
        nx.draw(G, node_color = node_colors, pos = pos)
    return {
        'G': G,
        'pos': pos,
        'node_colors': node_colors,
        'node_labels': node_labels,
    }



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import numpy as np
from ._utils import _softmax_class
from sklearn import metrics, datasets
from sklearn.naive_bayes import MultinomialNB
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.model_selection import train_test_split
from .stem import _classification_textcleaning_stemmer
from .texts._text_functions import separate_dataset
from ._models._sklearn_model import USER_BAYES
from ._utils._paths import PATH_SENTIMENTS, S3_PATH_SENTIMENTS
from .texts.vectorizer import SkipGramVectorizer


def available_sparse_deep_model():
    """
    List available sparse deep learning sentiment analysis models.
    """
    return ['fast-text-char']


def available_deep_model():
    """
    List available deep learning sentiment analysis models.
    """
    return [
        'fast-text',
        'hierarchical',
        'bahdanau',
        'luong',
        'bidirectional',
        'bert',
        'entity-network',
    ]


def sparse_deep_model(model = 'fast-text-char', validate = True):
    """
    Load deep learning sentiment analysis model.

    Parameters
    ----------
    model : str, optional (default='luong')
        Model architecture supported. Allowed values:

        * ``'fast-text-char'`` - Fast-text architecture for character based n-grams, embedded and logits layers only
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    SPARSE_SOFTMAX: malaya._models._tensorflow_model.SPARSE_SOFTMAX class
    """
    return _softmax_class.sparse_deep_model(
        PATH_SENTIMENTS,
        S3_PATH_SENTIMENTS,
        'sentiment',
        ['negative', 'positive'],
        2,
        model = model,
        validate = validate,
    )


def deep_model(model = 'luong', validate = True):
    """
    Load deep learning sentiment analysis model.

    Parameters
    ----------
    model : str, optional (default='luong')
        Model architecture supported. Allowed values:

        * ``'fast-text'`` - Fast-text architecture, embedded and logits layers only
        * ``'hierarchical'`` - LSTM with hierarchical attention architecture
        * ``'bahdanau'`` - LSTM with bahdanau attention architecture
        * ``'bidirectional'`` - LSTM with Bidirectional RNN architecture
        * ``'luong'`` - LSTM with luong attention architecture
        * ``'bert'`` - Deep Bidirectional transformers architecture
        * ``'entity-network'`` - Recurrent Entity-Network architecture
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    SOFTMAX: malaya._models._tensorflow_model.SOFTMAX class
    """
    return _softmax_class.deep_model(
        PATH_SENTIMENTS,
        S3_PATH_SENTIMENTS,
        'sentiment',
        ['negative', 'positive'],
        model = model,
        validate = validate,
    )


def multinomial(validate = True):
    """
    Load multinomial sentiment model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    USER_BAYES : malaya._models._sklearn_model.USER_BAYES class
    """
    return _softmax_class.multinomial(
        PATH_SENTIMENTS,
        S3_PATH_SENTIMENTS,
        'sentiment',
        ['negative', 'positive'],
        validate = validate,
    )


def xgb(validate = True):
    """
    Load XGB sentiment model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    USER_XGB : malaya._models._sklearn_model.USER_XGB class
    """
    return _softmax_class.xgb(
        PATH_SENTIMENTS,
        S3_PATH_SENTIMENTS,
        'sentiment',
        ['negative', 'positive'],
        validate = validate,
    )


def train_multinomial(corpus, vector = 'tfidf', split_size = 0.2, **kwargs):
    """
    Train a multinomial model using your own corpus / dataset.

    Parameters
    ----------
    corpus : list, str
        List of (string, label). Allowed tuple and list only.

    vector: str
        Vectorization techniques to feed into the model. Allowed values:

        * ``'tfidf'`` - TF-IDF
        * ``'bow'`` - bag-of-word
        * ``'skip-gram'`` - bag-of-word with skipping module

    split_size: float
        Split size to train and test. Should bigger than 0, less than 1.

    Returns
    -------
    USER_BAYES: malaya.sklearn_model.USER_BAYES class
    """
    if not (
        isinstance(corpus, str)
        or isinstance(corpus, list)
        or isinstance(corpus, tuple)
    ):
        raise ValueError(
            'corpus must be a string location or list of strings or tuple of strings'
        )
    if not isinstance(vector, str):
        raise ValueError('vector must be a string')
    if not isinstance(split_size, float):
        raise ValueError('split_size must be a float')
    if not (split_size > 0 and split_size < 1):
        raise ValueError('split_size must bigger than 0, less than 1')
    multinomial, labels, vectorize = None, None, None
    if isinstance(corpus, str):
        trainset = datasets.load_files(
            container_path = corpus, encoding = 'UTF-8'
        )
        trainset.data, trainset.target = separate_dataset(trainset)
        data, target = trainset.data, trainset.target
        labels = trainset.target_names
    if isinstance(corpus, list) or isinstance(corpus, tuple):
        if not len(corpus[0]) == 2:
            raise ValueError(
                'element of corpus must be list or tuple of (string, label)'
            )
        if not isinstance(corpus[0][0], str):
            raise ValueError('left hand side of element must be a string')
        corpus = np.array(corpus)
        data, target = corpus[:, 0].tolist(), corpus[:, 1].tolist()
        labels = np.unique(target).tolist()
        target = LabelEncoder().fit_transform(target)

    from sklearn.utils import shuffle

    data, target = shuffle(data, target)

    for i in range(len(data)):
        data[i] = _classification_textcleaning_stemmer(data[i])
    if 'tfidf' in vector.lower():
        vectorize = TfidfVectorizer(**kwargs).fit(data)
    elif 'bow' in vector.lower():
        vectorize = CountVectorizer(**kwargs).fit(data)
    elif 'skip-gram' in vector.lower():
        vectorize = SkipGramVectorizer(**kwargs).fit(data)
    else:
        raise Exception(
            "vectorizing techniques not supported, only support ['tf-idf', 'bow', 'skip-gram']"
        )
    vectors = vectorize.transform(data)
    multinomial = MultinomialNB()
    if split_size:
        train_X, test_X, train_Y, test_Y = train_test_split(
            vectors, target, test_size = split_size
        )
        multinomial.partial_fit(train_X, train_Y, classes = np.unique(target))
        predicted = multinomial.predict(test_X)
        print(
            metrics.classification_report(
                test_Y, predicted, target_names = labels
            )
        )
    else:
        multinomial.partial_fit(vectors, target, classes = np.unique(target))
        predicted = multinomial.predict(vectors)
        print(
            metrics.classification_report(
                target, predicted, target_names = labels
            )
        )
    return USER_BAYES(
        multinomial,
        labels,
        vectorize,
        cleaning = _classification_textcleaning_stemmer,
    )
